#include <string.h>
#include <WiFi.h>
#include <WiFiUdp.h>

#include <string>
#include <map>

#include <PZEM004Tv30.h>

#define PIN_OXCO 2
#define PIN_ADC  A0

#include <stdio.h>

#include "bertrik-sikken/statemachine.h"

#define NAME "MainsMeasurer"

#define SAMPLE_FREQUENCY    10000

#define UPLOAD_INTERVAL 5000

#define BUF_SIZE    2048
uint16_t buffer[BUF_SIZE];
volatile uint32_t bufr = 0;
volatile uint32_t bufw = 0;
volatile bool overflow = false;
volatile uint32_t sample_count = 0;

double last_freq_measurement = 0.;
bool   last_freq_measurement_updated = false;

uint32_t mqtt_reconnects = 0;

#include <PubSubClient.h>

std::string base_topic = "";

std::map<std::string, std::string> data;
mutex_t data_mutex;

WiFiClient   wclient;
PubSubClient mclient(wclient);

void sample_reset(void) {
	bufr = 0;
	bufw = 0;
	overflow = false;
}

void sampler() {
	uint32_t next = (bufw + 1) % BUF_SIZE;

	if (!overflow) {
		if (next != bufr) {
			buffer[bufw] = analogRead(PIN_ADC);
			bufw = next;
		}
		else {
			overflow = true;
		}
	}

	sample_count++;
}

bool sample_get(double *pval) {
	if (bufr == bufw)
		return false;

	uint32_t next = (bufr + 1) % BUF_SIZE;
	*pval = buffer[bufr];
	bufr = next;

	return true;
}

int compare_uint16(const void *v1, const void *v2) {
	uint16_t u1 = *((uint16_t *) v1);
	uint16_t u2 = *((uint16_t *) v2);

	return u1 - u2;
}

double do_freq() {
	// take stats
	sample_reset();

	while(!overflow)
		yield();

	qsort(buffer, BUF_SIZE, sizeof(uint16_t), compare_uint16);
	uint16_t q1  = buffer[2 * BUF_SIZE / 8];
	uint16_t med = buffer[4 * BUF_SIZE / 8];
	uint16_t q3  = buffer[6 * BUF_SIZE / 8];

	Serial.printf("stats: %u-%u-%u\r\n", q1, med, q3);

	// determine zero crossings
	sample_reset();
	StateMachine sm(q1 - med, q3 - med);
	int t = 0;
	uint32_t start = millis();
	double first = 0.0;
	double last = 0.0;
	int count = 0;
	bool done = false;
	while (!done && millis() - start < 3000) {
		double value = 0;
		if (sample_get(&value)) {
			double time = (double)t / SAMPLE_FREQUENCY;

			if (sm.process(time, value - med)) {
				switch (count) {
					case 0:
						first = sm.get_result();
						break;
					case 50:
						last = sm.get_result();
						done = true;
						break;
					default:
						break;
				}

				count++;
			}

			t++;
		}
	}

	double freq = 50.0 / (last - first);
	Serial.printf("n=%d,first=%f,last=%f,frequency=%f\r\n", count, first, last, freq);

	return freq;
}

PZEM004Tv30 *pzem = nullptr;

WiFiServer http_server(80);

void setup() {
	Serial.begin(115200);

	Serial.println(F(NAME " " __DATE__ " " __TIME__));

	watchdog_enable(5000, false);

	pinMode(1, OUTPUT);
	digitalWrite(1, HIGH);

	Serial.setDebugOutput(true);

	analogReadResolution(12);

	WiFi.hostname(NAME);

	WiFi.mode(WIFI_STA);

	auto cnt = WiFi.scanNetworks();

	for(int i=0; i<cnt; i++) {
		if (strcmp(WiFi.SSID(i), "www.vanheusden.com") == 0) {
			Serial.println(F("Attempting connection to thuis"));
			WiFi.begin("www.vanheusden.com", "hereasecretpassword");
		}
		else if (strcmp(WiFi.SSID(i), "revspace-pub-2.4ghz") == 0) {
			WiFi.begin("revspace-pub-2.4ghz", "");
			Serial.println(F("Attempting connection to revspace"));
		}
		else {
			Serial.print("Ignoring SSID: ");
			Serial.println(WiFi.SSID(i));
		}
	}

	while(WiFi.status() != WL_CONNECTED)
		delay(100);

	mclient.setServer("vps001.vanheusden.com", 1883);

        if (Serial2.setRX(5) && Serial2.setTX(4))
                Serial.println(F("Serial2 configured"));

	pzem = new PZEM004Tv30(Serial2);

	char buffer[2 * PICO_UNIQUE_BOARD_ID_SIZE_BYTES + 1] { 0 };
	pico_get_unique_board_id_string(buffer, sizeof buffer);

	base_topic = std::string(NAME) + "/" + std::string(buffer) + "/";

	mutex_init(&data_mutex);

	http_server.begin();

	pinMode(PIN_OXCO, INPUT);
	attachInterrupt(PIN_OXCO, sampler, RISING);

	digitalWrite(1, LOW);

	Serial.println(F("Go!"));
}

void setup1() {
}

void force_reset() {
	Serial.println(F("Reset"));

	watchdog_enable(1, 1);

	for(;;) {
	}
}

void MQTT_connect() {
	mclient.loop();

	if (!mclient.connected()) {
		mqtt_reconnects++;

		int tries = 0;

		while (!mclient.connected()) {
			Serial.println("Attempting MQTT connection... ");

			if (mclient.connect(NAME)) {
				Serial.println(F("Connected"));

				auto ip = WiFi.localIP();

				char buffer[32];
				snprintf(buffer, sizeof(buffer), "%d.%d.%d.%d", ip[0], ip[1], ip[2], ip[3]);

				std::string topic = base_topic + "stats/ip-address";
				mclient.publish(topic.c_str(), buffer);

				break;
			}

			watchdog_update();

			delay(500);

			if (++tries >= 20)
				force_reset();
		}
	}
}

volatile long int prev_upload = millis();

void loop() {
	// watchdog
	long int now = millis();

	if (now - prev_upload >= 300000l) {
		Serial.println(F(" *** watchdog ***"));
		force_reset();
	}

	WiFiClient client = http_server.available();

	if (client) {
		unsigned long start = millis();

		while(client.available() && millis() - start < 1000)
			client.read();

		client.println("HTTP/1.1 200 OK");
		client.println("Content-type:text/html");
		client.println("Connection: close");
		client.println();
		client.println("<!DOCTYPE html>");
		client.println("<html lang=\"en\">");
		client.println("<head>");
		client.println("<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">");
		client.println("<meta charset=\"utf-8\">");
		client.println("<link href=\"https://vanheusden.com/simple.min.css\" rel=\"stylesheet\" type=\"text/css\">");
		client.println("</head>");
		client.println("<body>");
		client.println("<h1>" NAME "</h1>");
		client.println("<article>");
		client.println("<header><h2>by folkert van heusden</h2></header>");
		client.println("<table border=1>");
		mutex_enter_blocking(&data_mutex);
		for(auto & it: data) {
			client.println("<tr>");

			std::string temp1 = "<td>" + it.first + "</td>";
			client.println(temp1.c_str());

			std::string temp2 = "<td>" + it.second + "</td>";
			client.println(temp2.c_str());

			client.println("</tr>");
		}
		mutex_exit(&data_mutex);
		client.println("</table>");
		client.println("</article>");
		client.println("</body>");
		client.println("</html>");

		client.stop();
	}
}

void publish_value(const std::string name, double value)
{
	char buffer[16] { 0 };
	snprintf(buffer, sizeof buffer, "%e", value);

	std::string topic = base_topic + name;

	mclient.publish(topic.c_str(), buffer);

	mutex_enter_blocking(&data_mutex);
	data.insert_or_assign(topic, buffer);
	mutex_exit(&data_mutex);

	Serial.print(topic.c_str());
	Serial.print(F(" - "));
	Serial.println(buffer);
}

void publish_value(const std::string name, int32_t value)
{
	char buffer[16] { 0 };
	snprintf(buffer, sizeof buffer, "%ld", value);

	std::string topic = base_topic + name;

	mclient.publish(topic.c_str(), buffer);

	mutex_enter_blocking(&data_mutex);
	data.insert_or_assign(topic, buffer);
	mutex_exit(&data_mutex);

	Serial.print(topic.c_str());
	Serial.print(F(" - "));
	Serial.println(buffer);
}

void publish_value(const std::string name, uint32_t value)
{
	char buffer[16] { 0 };
	snprintf(buffer, sizeof buffer, "%lu", value);

	std::string topic = base_topic + name;

	mclient.publish(topic.c_str(), buffer);

	mutex_enter_blocking(&data_mutex);
	data.insert_or_assign(topic, buffer);
	mutex_exit(&data_mutex);

	Serial.print(topic.c_str());
	Serial.print(F(" - "));
	Serial.println(buffer);
}

uint32_t n_measurements = 0;

void loop1() {
	MQTT_connect();

	watchdog_update();

	long int sc = sample_count;
	long int now = millis();
	if (now - prev_upload < 5000)
		return;

	digitalWrite(1, HIGH);

	prev_upload = now;

	// generic stats
	Serial.print(now);
	Serial.print(' ');
	Serial.print(sc);
	Serial.print(' ');
	long int drift = sc - now * 10;
	Serial.print(drift);
	Serial.println("");

	n_measurements++;

	publish_value("stats/uptime", int32_t(millis()));
	publish_value("stats/clock-drift", int32_t(drift));
	publish_value("stats/n-measurements", n_measurements);
	publish_value("stats/mqtt-reconnects", mqtt_reconnects);
	publish_value("stats/wifi-rssi", int32_t(WiFi.RSSI()));

	// net frequency via bertrik's code
	double frequency = do_freq();

	if (frequency >= 30. && frequency <= 70.)
		publish_value("BS-ZMPT101B/frequency", frequency);
	else
		publish_value("BS-ZMPT101B/frequency-fail", frequency);

	// pzem sensor
	char buffer[8] { 0 };
	snprintf(buffer, sizeof buffer, "%02x", pzem->readAddress());
	std::string topic = base_topic + "PZEM004Tv30/address";
	mclient.publish(topic.c_str(), buffer);

	int failure_count = 0;

	float voltage = pzem->voltage();
	if (!isnan(voltage))
		publish_value("PZEM004Tv30/voltage", voltage);
	else
		failure_count++;

	float current = pzem->current();
	if (!isnan(current))
		publish_value("PZEM004Tv30/current", current);
	else
		failure_count++;

	float power = pzem->power();
	if (!isnan(power))
		publish_value("PZEM004Tv30/power", power);
	else
		failure_count++;

	float energy = pzem->energy();
	if (!isnan(energy))
		publish_value("PZEM004Tv30/energy", energy);
	else
		failure_count++;

	float pzem_frequency = pzem->frequency();
	if (!isnan(pzem_frequency))
		publish_value("PZEM004Tv30/frequency", pzem_frequency);
	else
		failure_count++;

	float pf = pzem->pf();
	if (!isnan(pf))
		publish_value("PZEM004Tv30/power-factor", pf);
	else
		failure_count++;

	publish_value("stats/pzem-failed", int32_t(failure_count));

	digitalWrite(1, LOW);
}
